import sys
import numpy as np

# C-style range function.
def normalize_point(value, in_min, in_max, out_min, out_max):
	return (value - in_min) * (out_max - out_min) / (in_max - in_min) + out_min

# Grab user input - Input path and output path.
if len(sys.argv) < 2:
	print("Data Normalizer for ECSE-4750")
	print("Usage: {} <inputfile> <outputfile>".format(sys.argv[0]))
	exit(-1)

else:
	print("Input file: {}".format(sys.argv[1]))
	print("Outputting to: {}".format(sys.argv[2]))

	# Read the input file.
	with open(sys.argv[1], 'r') as f:
		data = f.readlines()
	raw_data_float = []

	# For each line of data, split the strings and convert to floats.
	for line in data:
		line = line.split()
		for coord in line:
			coord = float(coord)
		raw_data_float.append((float(line[0]),float(line[1]),float(line[2]),float(line[3])
							   ,float(line[4]),float(line[5])
							   ,float(line[6]),float(line[7]),float(line[8])))

	# Find the min and max of the data for normalizing it.
	float_min = 0.0
	float_max = 0.0
	for point in raw_data_float:
		if max(point) > float_max:
			float_max = max(point)
		if min(point) < float_min:
			float_min = min(point)

	# Go through and actually normalize each point.
	normalized_data_float = []
	normalized_data_normals = []

	print("Normalizing with these values:")
	print("Data Max: {}".format(float_max))
	print("Data Min: {}".format(float_min))
	print("Normalizing...")

	# Assemble three points as a triangle.
	for triangle in raw_data_float:
		normalized_triangle = tuple([normalize_point(x, -5.0, 5.0, -1, 1) for x in triangle])
		normalized_data_float.append(normalized_triangle)

	print("Complete.")
	print("Writing to file...")
	output_vertices = ""

	# Turn everything into a string for exporting into .js
	for pt in normalized_data_float:
		output_vertices += "\tvec3({}, {}, {}),\n".format(pt[0], pt[1], pt[2])
		output_vertices += "\tvec3({}, {}, {}),\n".format(pt[3], pt[4], pt[5])
		output_vertices += "\tvec3({}, {}, {}),\n".format(pt[6], pt[7], pt[8])

	# Strip off that final comma and newline
	output_vertices = output_vertices[:-2]

	# Write to a file.
	with open(sys.argv[2], 'wb') as f:
		f.write("var vertices = [\n".encode("ASCII"))
		f.write(output_vertices.encode("ASCII"))
		f.write("\n];\n".encode("ASCII"))
		f.write("var count = {};".format(len(normalized_data_float * 9)).encode("ASCII"))
	print("Done!")
